//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include "gnss/processor/processor_tracker_gnss.h"

#include "gnss/capture/capture_gnss.h"
#include "gnss/feature/feature_gnss_tdcp.h"
#include "gnss/feature/feature_gnss_fix.h"
#include "gnss/factor/factor_gnss_tdcp.h"
#include "gnss/factor/factor_gnss_tdcp_batch.h"
#include "gnss/factor/factor_gnss_pseudo_range.h"
#include "gnss/factor/factor_gnss_fix_3d.h"
#include "gnss_utils/utils/rcv_position.h"
#include "gnss_utils/tdcp.h"

namespace wolf
{

ProcessorTrackerGnss::ProcessorTrackerGnss(ParamsProcessorTrackerGnssPtr _params_tracker_gnss) :
        ProcessorTrackerFeature("ProcessorTrackerGnss", "PO", 3, _params_tracker_gnss),
        params_tracker_gnss_(_params_tracker_gnss),
        outliers_pseudorange_(0),
        outliers_tdcp_(0),
        inliers_pseudorange_(0),
        inliers_tdcp_(0),
        first_pos_(Eigen::Vector3d::Zero())
{
}

void ProcessorTrackerGnss::preProcess()
{
    WOLF_DEBUG("ProcessorTrackerGnss::preProcess");

    GnssUtils::SnapshotPtr inc_snapshot = std::static_pointer_cast<CaptureGnss>(incoming_ptr_)->getSnapshot();
    GnssUtils::Options copy_opt = params_tracker_gnss_->gnss_opt;
    bool eph_sbas34 = params_tracker_gnss_->gnss_opt.sateph == EPHOPT_SBAS3 or
                      params_tracker_gnss_->gnss_opt.sateph == EPHOPT_SBAS4;
    bool iono_sbas34 = params_tracker_gnss_->gnss_opt.ionoopt == IONOOPT_SBAS3 or
                       params_tracker_gnss_->gnss_opt.ionoopt == IONOOPT_SBAS4;

#ifdef _WOLF_DEBUG
    int n_initial = inc_snapshot->getObservations()->size();
    std::string initial_str;
    for (auto obs : inc_snapshot->getObservations()->getObservations())
        initial_str += std::to_string(obs.sat) + " ";
#endif
    // overload SBAS3 and SBAS4 first chance
    if (eph_sbas34)
        copy_opt.sateph = EPHOPT_SBAS;
    if (iono_sbas34)
        copy_opt.ionoopt = IONOOPT_SBAS;

    // compute satellites positions
    inc_snapshot ->computeSatellites(copy_opt.sateph);

    /* this fix is used for:
     *   - set ENU
     *   - compute azimuths and elevations
     *   - Take the eventual discarded sats by RAIM
     *   - initialize clock bias
     **/
    fix_incoming_ = GnssUtils::computePos(*inc_snapshot->getObservations(),
                                          *inc_snapshot->getNavigation(),
                                          params_tracker_gnss_->fix_opt);

    // Initialize clock bias stateblocks in capture
    if (fix_incoming_.success)
    {
        incoming_ptr_->getStateBlock('T') ->setState(Eigen::Vector1d(CLIGHT*fix_incoming_.rcv_bias(0)));
        if (sensor_gnss_->isStateBlockDynamic('G'))
            incoming_ptr_->getStateBlock('G')->setState(Eigen::Vector1d(CLIGHT*fix_incoming_.rcv_bias(1)));
        if (sensor_gnss_->isStateBlockDynamic('E'))
            incoming_ptr_->getStateBlock('E')->setState(Eigen::Vector1d(CLIGHT*fix_incoming_.rcv_bias(2)));
        if (sensor_gnss_->isStateBlockDynamic('M'))
            incoming_ptr_->getStateBlock('M')->setState(Eigen::Vector1d(CLIGHT*fix_incoming_.rcv_bias(3)));

        if (first_pos_.squaredNorm() < 1e-6)
            first_pos_ = fix_incoming_.pos;
    }

    // Set ECEF-ENU
    if (!sensor_gnss_->isEnuDefined() and sensor_gnss_->isEnuModeAuto() )
    {
        if (fix_incoming_.success)
        {
            WOLF_DEBUG("setting ECEF-ENU: ", fix_incoming_.pos.transpose());
            sensor_gnss_->setEcefEnu(fix_incoming_.pos, true);
        }
        else
            WOLF_WARN("Failed to compute fix to set ECEF-ENU in auto mode. Message: ", fix_incoming_.msg);
    }
    // Fix ENU-MAP
    if ((fix_incoming_.pos - first_pos_).norm() > params_tracker_gnss_->enu_map_fix_dist)
    {
        sensor_gnss_->getEnuMapTranslation()->fix();
        sensor_gnss_->getEnuMapRoll()->fix();
        sensor_gnss_->getEnuMapPitch()->fix();
        sensor_gnss_->getEnuMapYaw()->fix();
    }

    WOLF_DEBUG("TS: ", incoming_ptr_->getTimeStamp(), " - Fix solution (ECEF): ", fix_incoming_.pos.transpose(), " - Fix solution (GEO): ", fix_incoming_.lat_lon.transpose());

    // store initial observation before filtering
    GnssUtils::ObservationsPtr init_obs;
    if (eph_sbas34 or iono_sbas34)
        init_obs = std::make_shared<GnssUtils::Observations>(*inc_snapshot->getObservations());

    // filter observations (available ephemeris, constellations and elevation&SNR)
    #ifdef _WOLF_DEBUG
        auto discarded_gnssutils =
    #endif
    inc_snapshot->filterObservations(fix_incoming_.discarded_sats, // discarded sats
                                     fix_incoming_.sat_azel,
                                     false, // check code
                                     false, // check carrier phase
                                     copy_opt);

    // compute corrected Ranges
    inc_snapshot->computeRanges(fix_incoming_.sat_azel,
                                fix_incoming_.lat_lon,
                                copy_opt);

    /* NOT ENOUGH SATS/RANGES in case of:
     *     eph:  EPHOPT_SBAS3 or EPHOPT_SBAS4
     *     iono: IONOOPT_SBAS3 or IONOOPT_SBAS4
     */
    if (inc_snapshot->getRanges().size() < params_tracker_gnss_->min_sbas_sats and
        (eph_sbas34 or iono_sbas34))
    {
        WOLF_DEBUG("ProcessorTrackerGnss::preProcess: with SBAS3/SBAS4, not enough ranges: ", inc_snapshot->getObservations()->size(), " Computing with SBAS2/BRDC");

        if (params_tracker_gnss_->gnss_opt.sateph == EPHOPT_SBAS3)
            copy_opt.sateph = EPHOPT_SBAS2;
        if (params_tracker_gnss_->gnss_opt.sateph == EPHOPT_SBAS4)
            copy_opt.sateph = EPHOPT_BRDC;

        if (params_tracker_gnss_->gnss_opt.ionoopt == IONOOPT_SBAS3)
            copy_opt.ionoopt = IONOOPT_SBAS2;
        if (params_tracker_gnss_->gnss_opt.ionoopt == IONOOPT_SBAS4)
            copy_opt.ionoopt = IONOOPT_BRDC;

        // reset observations
        inc_snapshot->setObservations(init_obs);
        inc_snapshot->getRanges().clear();
        // recompute satellites
        inc_snapshot ->computeSatellites(copy_opt.sateph);
        // filter
        #ifdef _WOLF_DEBUG
            discarded_gnssutils =
        #endif
        inc_snapshot->filterObservations(fix_incoming_.discarded_sats, // discarded sats
                                         fix_incoming_.sat_azel,
                                         false, // check code
                                         false, // check carrier phase
                                         copy_opt);


        // recompute corrected Ranges
        inc_snapshot->computeRanges(fix_incoming_.sat_azel,
                                    fix_incoming_.lat_lon,
                                    copy_opt);

        WOLF_DEBUG("ProcessorTrackerGnss::preProcess: with SBAS2/BRDC, obtained ranges: ", inc_snapshot->getRanges().size());
    }

    // create features pseudorange
    for (auto obs_sat : inc_snapshot->getObservations()->getObservations())
    {
        if(inc_snapshot->getSatellites().count(obs_sat.sat) == 0)
        {
            WOLF_DEBUG("ProcessorTrackerGnss::preProcess(): Creating features of obs_sat ", obs_sat.sat , " satellite not found");
            continue;
        }
        if (inc_snapshot->getRanges().count(obs_sat.sat) == 0)
        {
            WOLF_DEBUG("ProcessorTrackerGnss::preProcess(): Creating features of obs_sat ", obs_sat.sat , " pseudorange not found");
            continue;
        }

        auto feat = FeatureBase::emplace<FeatureGnssSatellite>(incoming_ptr_, obs_sat, inc_snapshot->getSatellites().at(obs_sat.sat), inc_snapshot->getRanges().at(obs_sat.sat));
        assert(untracked_incoming_features_.count(feat->satNumber()) == 0);
        untracked_incoming_features_[feat->satNumber()] = feat;
    }

#ifdef _WOLF_DEBUG
    std::string used_rtklib_str, discarded_rtklib_str, detected_str, discarded_gnssutils_str;
    for (auto sat : fix_incoming_.used_sats)
        used_rtklib_str += std::to_string(sat) + " ";
    for (auto sat : fix_incoming_.discarded_sats)
        discarded_rtklib_str += std::to_string(sat) + " ";
    for (auto ftr_pair : untracked_incoming_features_)
        detected_str += std::to_string(ftr_pair.first) + " ";
    for (auto sat_disc : discarded_gnssutils)
        if (fix_incoming_.discarded_sats.count(sat_disc) == 0)
            discarded_gnssutils_str += std::to_string(sat_disc) + " ";
#endif

    WOLF_DEBUG("ProcessorTrackerGnss::preProcess()",
              "\n\tinitial observations: ", n_initial, " (", initial_str, ")",
              "\n\tRTKLIB used: ", fix_incoming_.used_sats.size(), " (", used_rtklib_str, ")",
              //"\n\tRTKLIB discarded: ", fix_incoming_.discarded_sats.size(), " (", discarded_rtklib_str, ")",
              "\n\tgnssutils discarded: ", n_initial - untracked_incoming_features_.size() - fix_incoming_.discarded_sats.size(), " (", discarded_gnssutils_str, ")",
              "\n\tdetected incoming features: ", untracked_incoming_features_.size(), " (", detected_str, ")");
}

unsigned int ProcessorTrackerGnss::trackFeatures(const FeatureBasePtrList& _features_in,
                                                 const CaptureBasePtr& _capture,
                                                 FeatureBasePtrList& _features_out,
                                                 FeatureMatchMap& _feature_correspondences)
{
    if (_features_in.empty())
        return 0;

    WOLF_DEBUG_COND(_features_out.begin() == new_features_incoming_.begin(), "ProcessorTrackerGnss::trackFeatures tracking " , _features_in.size() , " new features...");
    WOLF_DEBUG_COND(_features_out.begin() == known_features_incoming_.begin(), "ProcessorTrackerGnss::trackFeatures tracking " , _features_in.size() , " known features...");

    assert(_capture == incoming_ptr_);

    int common_sats = 0;

    for (auto feat_in : _features_in)
    {
        auto feat_in_gnss = std::dynamic_pointer_cast<FeatureGnssSatellite>(feat_in);
        assert(feat_in_gnss);

        int sat_num = std::static_pointer_cast<FeatureGnssSatellite>(feat_in)->satNumber();

        WOLF_DEBUG("tracking " , feat_in->trackId() , ", sat number ", sat_num);

        if (untracked_incoming_features_.count(sat_num) != 0)
            common_sats++;

        if (untracked_incoming_features_.count(sat_num) != 0 and
            std::abs(untracked_incoming_features_.at(sat_num)->getObservation().L[0]) > 1e-12) // Track only carrier phase valid
        {
            auto ftr = untracked_incoming_features_[sat_num];
            untracked_incoming_features_.erase(sat_num);
            _features_out.push_back(ftr);
            _feature_correspondences[ftr] = std::make_shared<FeatureMatch>(FeatureMatch({feat_in,0}));

            WOLF_DEBUG("\ttracked: " , feat_in->trackId() , " last: " , feat_in->id() , " inc: " , ftr->id() , " !" );
        }
        else
        {
            WOLF_DEBUG_COND(untracked_incoming_features_.count(sat_num) == 0, "\tnot tracked, missing satellite" );
            WOLF_DEBUG_COND(untracked_incoming_features_.count(sat_num) == 1, "\tnot tracked, wrong carrier phase value" );
        }
    }
    WOLF_WARN_COND(_features_out.empty() and
                   _features_out.begin() == known_features_incoming_.begin(),
                   "ProcessorTrackerGnss::trackFeatures: LOST TRACK OF ALL SATELLITES of ", _features_in.size(),
                   " - ", common_sats, " due to wrong Carrier Phase data.");
    WOLF_DEBUG_COND(_features_out.begin() == new_features_incoming_.begin(),
                    "ProcessorTrackerGnss::trackFeatures: tracked " ,
                    _features_out.size(), " new features (of ", _features_in.size(), ")");
    WOLF_DEBUG_COND(_features_out.begin() == known_features_incoming_.begin(),
                    "ProcessorTrackerGnss::trackFeatures: tracked " ,
                    _features_out.size(), " known features (of ", _features_in.size(), ")");

    return _features_out.size();
}

bool ProcessorTrackerGnss::voteForKeyFrame() const
{
    //WOLF_DEBUG("ProcessorTrackerGnss::voteForKeyFrame");

    // too old origin
    if (origin_ptr_== nullptr or (last_ptr_->getTimeStamp() - origin_ptr_->getTimeStamp()) > params_tracker_gnss_->max_time_span )
    {
        WOLF_DEBUG( "Vote for KF because of time span or null origin" );
        return true;
    }

    /* KNOWN FEATURES
     * If tracked sats are lower than minimum (min_features_for_keyframe), a KF in last will be useful
     * if it allows to add more satellites (untracked features in last is not empty)
     */
    WOLF_DEBUG("Nbr. of active feature tracks: " , known_features_incoming_.size() );
    if (known_features_incoming_.size() < params_tracker_feature_->min_features_for_keyframe
        and not untracked_last_features_.empty() )
    {
        WOLF_DEBUG( "Vote for KF because of too less known_features_incoming and not empty untracked in last" );
        return true;
    }

    // TODO: more alternatives?

   // otherwise
   return false;
}

unsigned int ProcessorTrackerGnss::detectNewFeatures(const int& _max_new_features,
                                                     const CaptureBasePtr& _capture,
                                                     FeatureBasePtrList& _features_out)
{
    WOLF_DEBUG("ProcessorTrackerGnss::detectNewFeatures");

    assert(_capture == last_ptr_);

    for (auto feat_pair : untracked_last_features_)
    {
        if (_features_out.size() == _max_new_features)
            break;

        // discard non-valid carrier phase
        if (std::abs(std::static_pointer_cast<FeatureGnssSatellite>(feat_pair.second)->getObservation().L[0]) < 1e-12)
            continue;

        _features_out.push_back(feat_pair.second);
        WOLF_DEBUG("feature " , feat_pair.second->id() , " detected!" );
    }
    WOLF_DEBUG(_features_out.size() , " new features detected!");

    return _features_out.size();
}

void ProcessorTrackerGnss::establishFactors()
{
    WOLF_DEBUG("ProcessorTrackerGnss::establishFactors");

    // initialize frame state with antenna position in map coordinates
    // (since we don't have orientation for removing extrinsics)
    if (params_tracker_gnss_->init_frames and fix_last_.success)
        last_ptr_->getFrame()->getP()->setState(sensor_gnss_->getREnuMap().transpose() * (sensor_gnss_->getREnuEcef() * fix_last_.pos + sensor_gnss_->gettEnuEcef() - sensor_gnss_->gettEnuMap()));

    FactorBasePtrList new_factors;

    // FIX FACTOR
    if (params_tracker_gnss_->fix)
    {
        GnssUtils::SnapshotPtr last_snapshot = std::static_pointer_cast<CaptureGnss>(last_ptr_)->getSnapshot();


        auto fix = GnssUtils::computePos(*last_snapshot->getObservations(),
                                          *last_snapshot->getNavigation(),
                                          params_tracker_gnss_->gnss_opt);
        if (fix.success)
        {
            FeatureBasePtr ftr = FeatureBase::emplace<FeatureGnssFix>(last_ptr_, fix);

            FactorBase::emplace<FactorGnssFix3d>(ftr,
                                                 ftr,
                                                 sensor_gnss_,
                                                 shared_from_this(),
                                                 params_tracker_gnss_->apply_loss_function,
                                                 FAC_ACTIVE);
        }
    }

    // PSEUDO RANGE FACTORS (all sats)
    if (params_tracker_gnss_->pseudo_ranges)
        for (auto ftr : last_ptr_->getFeatureList())
        {
            auto ftr_sat = std::dynamic_pointer_cast<FeatureGnssSatellite>(ftr);
            if (ftr_sat == nullptr)
                continue;

            // Check valid measurement
            if (std::abs(ftr_sat->getMeasurement()(0)) < 1e-12)
            {
                WOLF_WARN("Ignoring feature with not valid pseudorange: ", ftr_sat->getMeasurement()(0));
                continue;
            }

            // unfix clock bias
            last_ptr_->getStateBlock('T')->unfix();
            // unfix clock bias inter-constellation (if observed)
            if (ftr_sat->getSatellite().sys == SYS_GLO)
                last_ptr_->getStateBlock('G')->unfix();
            if (ftr_sat->getSatellite().sys == SYS_GAL)
                last_ptr_->getStateBlock('E')->unfix();
            if (ftr_sat->getSatellite().sys == SYS_CMP)
                last_ptr_->getStateBlock('M')->unfix();

            // Emplace a pseudo range factor
            auto new_fac = FactorBase::emplace<FactorGnssPseudoRange>(ftr_sat,
                                                                      ftr_sat,
                                                                      sensor_gnss_,
                                                                      shared_from_this(),
                                                                      params_->apply_loss_function);
            new_factors.push_back(new_fac);

            WOLF_DEBUG("FactorGnssPseudoRange emplaced for satellite: ", ftr_sat->satNumber());
        }

    // TDCP FACTORS (tracked sats)
    if (origin_ptr_ != last_ptr_ and params_tracker_gnss_->tdcp_enabled)
    {
        // Displacement factor from batch TDCP (FactorGnssTdcp3d)
        if (params_tracker_gnss_->tdcp_params.batch)
        {
            WOLF_DEBUG("TDCP BATCH frame ", last_ptr_->getFrame()->id());
            FactorBasePtr last_fac_ptr = nullptr;

            auto frame_map = getProblem()->getTrajectory()->getFrameMap();
            for (auto frame_rev_iter = frame_map.rbegin();
                frame_rev_iter != frame_map.rend();
                ++frame_rev_iter)
            {
                auto ref_KF = frame_rev_iter->second;
                WOLF_DEBUG("TDCP BATCH ref frame ", ref_KF->id());

                // discard non-key frames, last-last pair and frames without CaptureGnss
                if (ref_KF == last_ptr_->getFrame() or
                    ref_KF->getCaptureOf(getSensor(),"CaptureGnss") == nullptr)
                    continue;

                // static cast
                auto ref_cap_gnss = std::static_pointer_cast<CaptureGnss>(ref_KF->getCaptureOf(getSensor(),"CaptureGnss"));
                auto last_cap_gnss = std::static_pointer_cast<CaptureGnss>(last_ptr_);

                // dt
                double dt = last_ptr_->getTimeStamp() - ref_cap_gnss->getTimeStamp();
                WOLF_DEBUG("TDCP BATCH dt = ", dt);

                // discard strange cases
                if (dt <= 0)
                    continue;

                // within time window
                if (dt > params_tracker_gnss_->tdcp_params.time_window)
                    break;

                // discard removing Frame/capture/feature
                if (ref_cap_gnss->isRemoving() or ref_cap_gnss->getFrame()->isRemoving())
                    continue;

                // get common sats from tracking
                auto matches = track_matrix_.matches(ref_cap_gnss, last_cap_gnss);
                WOLF_DEBUG("TDCP BATCH matches ", matches.size());
                std::set<int> common_sats;
                for (auto match : matches)
                {
                    assert(std::dynamic_pointer_cast<FeatureGnssSatellite>(match.second.first)->satNumber());
                    assert(std::dynamic_pointer_cast<FeatureGnssSatellite>(match.second.second)->satNumber());
                    assert(std::static_pointer_cast<FeatureGnssSatellite>(match.second.first)->satNumber() ==
                           std::static_pointer_cast<FeatureGnssSatellite>(match.second.second)->satNumber());
                    common_sats.insert(std::static_pointer_cast<FeatureGnssSatellite>(match.second.first)->satNumber());
                }
                WOLF_DEBUG("TDCP BATCH common_sats: ", common_sats.size());
                //for (auto sat : common_sats)
                //    std::cout << sat << " ";
                //std::cout << std::endl;

                // DEBUG: FIND COMMON SATELLITES OBSERVATIONS
                //std::set<int> common_sats_debug = GnssUtils::Observations::findCommonObservations(*ref_cap_gnss->getSnapshot()->getObservations(),
                //                                                                                  *last_cap_gnss->getSnapshot()->getObservations());
                //WOLF_DEBUG("TDCP BATCH common_sats_debug: ", common_sats_debug.size());
                //for (auto sat : common_sats_debug)
                //    std::cout << sat << " ";
                //std::cout << std::endl;

                // reference ECEF position
                Eigen::Vector3d x_r = sensor_gnss_->computeFrameAntennaPosEcef(ref_KF);

                // compute TDCP batch
                auto tdcp_output = GnssUtils::Tdcp(ref_cap_gnss->getSnapshot(),
                                                   last_cap_gnss->getSnapshot(),
                                                   x_r,
                                                   common_sats,
                                                   Eigen::Vector4d::Zero(),
                                                   params_tracker_gnss_->tdcp_params);
                if (tdcp_output.success)
                {
                    WOLF_DEBUG("TDCP BATCH d = ", tdcp_output.d.transpose());
                    WOLF_DEBUG("TDCP BATCH cov =\n", tdcp_output.cov_d);

                    // EMPLACE FEATURE
                    auto factor_status = FAC_ACTIVE;
                    if (params_tracker_gnss_->tdcp_structure == "first-all")
                        factor_status = FAC_INACTIVE;

                    auto ftr = FeatureBase::emplace<FeatureGnssTdcp>(last_cap_gnss,
                                                                     tdcp_output.d,
                                                                     tdcp_output.cov_d);

                    // EMPLACE FACTOR
                    last_fac_ptr = FactorBase::emplace<FactorGnssTdcpBatch>(ftr,
                                                                            ftr,
                                                                            ref_cap_gnss,
                                                                            sensor_gnss_,
                                                                            shared_from_this(),
                                                                            params_->apply_loss_function,
                                                                            factor_status);

                    WOLF_DEBUG("TDCP BATCH factor ", last_fac_ptr->id() , " emplaced as ", (last_fac_ptr->getStatus() == FAC_ACTIVE ? "ACTIVE" : "INACTIVE"));

                    // only first (with previous frame) factor in "consecutive" structure
                    if (params_tracker_gnss_->tdcp_structure == "consecutive")
                        break;
                }
                else
                {
                    WOLF_DEBUG("TDCP BATCH failed with msg: ", tdcp_output.msg);
                }
            }
            // only last (with first frame) factor in "first-all" structure
            if (params_tracker_gnss_->tdcp_structure == "first-all" and last_fac_ptr)
            {
                last_fac_ptr->setStatus(FAC_ACTIVE);
                WOLF_DEBUG("TDCP BATCH factor ", last_fac_ptr->id() , " changed to ", (last_fac_ptr->getStatus() == FAC_ACTIVE ? "ACTIVE" : "INACTIVE"));
            }
        }
        // FACTOR per SATELLITE (FactorGnssTdcp)
        else
        {
            // iterate over tracked features of last
            for (auto ftr_k_pair : track_matrix_.snapshot(last_ptr_))
            {
                // current feature
                auto ftr_k = std::dynamic_pointer_cast<FeatureGnssSatellite>(ftr_k_pair.second);
                if(not ftr_k)
                    continue;

                // check valid measurement
                assert(std::abs(ftr_k->getObservation().L[0]) > 1e-12);

                WOLF_DEBUG("FeatureGnssSatellite sat: ", ftr_k->satNumber(), " id: ", ftr_k->id());

                // unfix clock bias
                last_ptr_->getStateBlock('T')->unfix();

                // emplace a tdcp factor from last to each KF
                auto ts_ftr_r_map = track_matrix_.trackAtKeyframes(ftr_k_pair.first);
                for (auto ts_ftr_r_it = ts_ftr_r_map.rbegin(); ts_ftr_r_it != ts_ftr_r_map.rend(); ts_ftr_r_it++)
                {
                    // cast reference feature
                    auto ftr_r = std::dynamic_pointer_cast<FeatureGnssSatellite>(ts_ftr_r_it->second);
                    if(not ftr_r)
                        continue;

                    // dt
                    double dt = ftr_k->getCapture()->getTimeStamp() - ts_ftr_r_it->first;

                    // discard incomming-last and last-last
                    if (dt < 0 or ftr_k == ftr_r)
                        continue;

                    // within time window
                    if (dt > params_tracker_gnss_->tdcp_params.time_window)
                        break;

                    // discard removing Frame/capture/feature
                    if (ftr_r->isRemoving() or ftr_r->getCapture()->isRemoving() or ftr_r->getCapture()->getFrame()->isRemoving())
                        continue;

                    // check valid measurement
                    assert(std::abs(ftr_r->getObservation().L[0]) > 1e-12);

                    WOLF_DEBUG("previous feature at KF: ", ftr_r->getCapture()->getFrame()->id(), " sat: ", ftr_r->satNumber(), " id: ", ftr_r->id());

                    // emplace tdcp factor
                    auto factor_status = FAC_ACTIVE;
                    if (params_tracker_gnss_->tdcp_structure == "first-all")
                        factor_status = FAC_INACTIVE;
                    double var_tdcp = dt * std::pow(params_tracker_gnss_->tdcp_params.sigma_atm,2) + std::pow(params_tracker_gnss_->tdcp_params.sigma_carrier,2);
                    auto new_fac = FactorBase::emplace<FactorGnssTdcp>(ftr_k,
                                                                       sqrt(var_tdcp),
                                                                       ftr_r,
                                                                       ftr_k,
                                                                       sensor_gnss_,
                                                                       shared_from_this(),
                                                                       params_tracker_gnss_->tdcp_params.loss_function,
                                                                       factor_status);
                    new_factors.push_back(new_fac);

                    // WOLF_DEBUG( "Factor: track: " , feature_in_last->trackId(),
                    //             " origin: "       , feature_in_origin->id() ,
                    //             " from last: "    , feature_in_last->id() );

                    // only first (with previous frame) factor in "consecutive" structure
                    if (params_tracker_gnss_->tdcp_structure == "consecutive")
                        break;
                }
                // only last (with first frame) factor in "first-all" structure
                if (params_tracker_gnss_->tdcp_structure == "first-all")
                    new_factors.back()->setStatus(FAC_ACTIVE);

                WOLF_DEBUG("All TDCP factors emplaced!");
            }
        }
    }

    // remove outliers
    if (!new_factors.empty() and (params_tracker_gnss_->remove_outliers or params_tracker_gnss_->remove_outliers_tdcp))
        removeOutliers(new_factors, last_ptr_);

    //getProblem()->print(4, 1, 1, 1);

    WOLF_DEBUG("ProcessorTrackerGnss::establishFactors done");
}

void ProcessorTrackerGnss::advanceDerived()
{
    ProcessorTrackerFeature::advanceDerived();

    untracked_last_features_ = std::move(untracked_incoming_features_);
    fix_last_ = fix_incoming_;
}

void ProcessorTrackerGnss::resetDerived()
{
    ProcessorTrackerFeature::resetDerived();

    untracked_last_features_ = std::move(untracked_incoming_features_);
    fix_last_ = fix_incoming_;
}

void ProcessorTrackerGnss::postProcess()
{
    // Fix orientation of last frame if no other processors
    if (getSensor()->getHardware()->getSensorList().size() == 1)
        getProblem()->getLastFrame()->getO()->fix();
}

void ProcessorTrackerGnss::removeOutliers(FactorBasePtrList fac_list, CaptureBasePtr cap)
{
    WOLF_DEBUG("ProcessorTrackerGnss::removeOutliers");

    FactorBasePtrList remove_fac;

    //WOLF_DEBUG( "PR ", getName()," rejectOutlier...");

    // PseudoRange states
    Eigen::Vector3d x;
    Eigen::Vector4d o;
    Eigen::Vector1d clock_bias;
    Eigen::Vector1d clock_bias_glo;
    Eigen::Vector1d clock_bias_gal;
    Eigen::Vector1d clock_bias_cmp;
    Eigen::Vector3d x_antena_pr;


    // state for pseudoranges is fix solution if OK
    if (cap == last_ptr_ and fix_last_.stat != 0 and params_tracker_gnss_->remove_outliers_with_fix)
    {
        WOLF_DEBUG("OUTLIERS COMPUTED USING fix_last");
        x = sensor_gnss_->getREnuMap().transpose() * (sensor_gnss_->getREnuEcef() * fix_last_.pos + sensor_gnss_->gettEnuEcef() - sensor_gnss_->gettEnuMap());
        o << 0,0,0,1;
        clock_bias <<  CLIGHT * fix_last_.rcv_bias(0);
        clock_bias_glo << CLIGHT * fix_last_.rcv_bias(1);
        clock_bias_gal << CLIGHT * fix_last_.rcv_bias(2);
        clock_bias_cmp << CLIGHT * fix_last_.rcv_bias(3);
        x_antena_pr = Eigen::Vector3d::Zero();

        //std::cout << "x:              " << x.transpose() << std::endl;
        //std::cout << "o:              " << o.transpose() << std::endl;
        //std::cout << "clock_bias:     " << clock_bias << std::endl;
        //std::cout << "clock_bias_glo: " << clock_bias_glo << std::endl;
        //std::cout << "clock_bias_gal: " << clock_bias_gal << std::endl;
        //std::cout << "clock_bias_cmp: " << clock_bias_cmp << std::endl;
        //std::cout << "x_antena_pr:    " << x_antena_pr.transpose() << std::endl;
        //std::cout << "frame p:        " << cap->getFrame()->getP()->getState().transpose() << std::endl;
        //std::cout << "frame o:        " << cap->getFrame()->getO()->getState().transpose() << std::endl;
        //std::cout << "sb clock:       " << cap->getStateBlock('T')->getState() << std::endl;
        //std::cout << "sb clock glo:   " << cap->getStateBlock('G')->getState() << std::endl;
        //std::cout << "sb clock gal:   " << cap->getStateBlock('E')->getState() << std::endl;
        //std::cout << "sb clock cmp:   " << cap->getStateBlock('M')->getState() << std::endl;
        //std::cout << "sb antena:      " << sensor_gnss_->getP()->getState().transpose() << std::endl;
    }
    else
    {
        x = cap->getFrame()->getP()->getState();
        o = cap->getFrame()->getO()->getState();
        clock_bias = cap->getStateBlock('T')->getState();
        clock_bias_glo = cap->getStateBlock('G')->getState();
        clock_bias_gal = cap->getStateBlock('E')->getState();
        clock_bias_cmp = cap->getStateBlock('M')->getState();
        x_antena_pr = sensor_gnss_->getP()->getState();
    }

    // TDCP states
    Eigen::Vector3d x_k(cap->getFrame()->getP()->getState());
    Eigen::Vector4d o_k(cap->getFrame()->getO()->getState());
    Eigen::Vector3d x_r(cap->getFrame()->getP()->getState());
    Eigen::Vector4d o_r(cap->getFrame()->getO()->getState());
    Eigen::Vector1d clock_bias_k(cap->getStateBlock('T')->getState());
    Eigen::Vector1d clock_bias_r(cap->getStateBlock('T')->getState());
    Eigen::Vector3d x_antena(sensor_gnss_->getP()->getState());

    // Common states
    Eigen::Vector3d t_ENU_map(sensor_gnss_->getEnuMapTranslation()->getState());
    Eigen::Vector1d roll_ENU_map(sensor_gnss_->getEnuMapRoll()->getState());
    Eigen::Vector1d pitch_ENU_map(sensor_gnss_->getEnuMapPitch()->getState());
    Eigen::Vector1d yaw_ENU_map(sensor_gnss_->getEnuMapYaw()->getState());

    //std::cout << "t_ENU_map:     " << t_ENU_map.transpose() << std::endl;
    //std::cout << "roll_ENU_map:  " << roll_ENU_map << std::endl;
    //std::cout << "pitch_ENU_map: " << pitch_ENU_map << std::endl;
    //std::cout << "yaw_ENU_map:   " << yaw_ENU_map << std::endl;

    // create double* array of a copy of the state for pseudo range factors
    double* parameters_glo_pr[9] = {x.data(),
                                    o.data(),
                                    clock_bias.data(),
                                    clock_bias_glo.data(),
                                    x_antena_pr.data(),
                                    t_ENU_map.data(),
                                    roll_ENU_map.data(),
                                    pitch_ENU_map.data(),
                                    yaw_ENU_map.data()};
    double* parameters_gal_pr[9] = {x.data(),
                                    o.data(),
                                    clock_bias.data(),
                                    clock_bias_gal.data(),
                                    x_antena_pr.data(),
                                    t_ENU_map.data(),
                                    roll_ENU_map.data(),
                                    pitch_ENU_map.data(),
                                    yaw_ENU_map.data()};
    double* parameters_cmp_pr[9] = {x.data(),
                                    o.data(),
                                    clock_bias.data(),
                                    clock_bias_cmp.data(),
                                    x_antena_pr.data(),
                                    t_ENU_map.data(),
                                    roll_ENU_map.data(),
                                    pitch_ENU_map.data(),
                                    yaw_ENU_map.data()};
    double* parameters_tdcp[11] = {x_r.data(),
                                   o_r.data(),
                                   clock_bias_r.data(),
                                   x_k.data(),
                                   o_k.data(),
                                   clock_bias_k.data(),
                                   x_antena.data(),
                                   t_ENU_map.data(),
                                   roll_ENU_map.data(),
                                   pitch_ENU_map.data(),
                                   yaw_ENU_map.data()};

    // create residuals
    double residual;

    for (auto fac : fac_list)
    {
        // PSEUDO RANGE FACTORS
        auto fac_pr = std::dynamic_pointer_cast<FactorGnssPseudoRange>(fac);
        auto ftr_sat = std::static_pointer_cast<FeatureGnssSatellite>(fac->getFeature());
        int sat = ftr_sat->satNumber();

        if (fac_pr)
        {
            // evaluate the factor in this state
            switch (ftr_sat->getSatellite().sys)
            {
                case SYS_GLO:
                    fac_pr->evaluate(parameters_glo_pr, &residual, nullptr);
                    break;
                case SYS_GAL:
                    fac_pr->evaluate(parameters_gal_pr, &residual, nullptr);
                    break;
                case SYS_CMP:
                    fac_pr->evaluate(parameters_cmp_pr, &residual, nullptr);
                    break;
                default:
                    fac_pr->evaluate(parameters_glo_pr, &residual, nullptr);
                    break;
            }

            // RTKLIB FIX error
            //int sat = std::static_pointer_cast<FeatureGnssSatellite>(fac->getFeature())->getSatellite().sat;
            //assert(fix_last_.prange_residuals.count(sat) && "sat not used when computing fix!");
            //WOLF_DEBUG("FactorGnssPseudoRange error = ", fac->getMeasurementSquareRootInformationUpper().inverse()*residual);
            //WOLF_DEBUG("RTKLIB pntpos error         = ", fix_last_.prange_residuals.at(sat));

            // discard if residual too high evaluated at the current estimation
            if (std::abs(residual) > params_tracker_gnss_->outlier_residual_th)
            {
                WOLF_WARN("Discarding FactorGnssPseudoRange, considered OUTLIER");
                WOLF_TRACE("Feature: ", fac->getMeasurement(),"\nError: ",fac->getMeasurementSquareRootInformationUpper().inverse()*residual);
                remove_fac.push_back(fac_pr);
                // store for statistics
                outliers_pseudorange_++;
                if (not sat_outliers_.count(sat))
                    sat_outliers_.emplace(sat, 1);
                else
                    sat_outliers_[sat]++;

            }
            else
                inliers_pseudorange_++;
        }
        // TDCP FACTORS
        auto fac_tdcp = std::dynamic_pointer_cast<FactorGnssTdcp>(fac);
        if (fac_tdcp and params_tracker_gnss_->remove_outliers_tdcp)
        {
            // update ref frame
            x_r = fac_tdcp->getCaptureOther()->getFrame()->getP()->getState();
            o_r = fac_tdcp->getCaptureOther()->getFrame()->getO()->getState();
            clock_bias_r = fac_tdcp->getCaptureOther()->getStateBlock('T')->getState();
            parameters_tdcp[0] = x_r.data();
            parameters_tdcp[1] = o_r.data();
            parameters_tdcp[2] = clock_bias_r.data();

            // evaluate
            fac_tdcp->evaluate(parameters_tdcp, &residual, nullptr);

            //WOLF_DEBUG("FactorGnssTdcp with residual = ", residual);

            // discard if residual too high evaluated at the current estimation
            if (std::abs(residual) > params_tracker_gnss_->outlier_residual_th)
            {
                WOLF_WARN("Discarding FactorGnssTdcp, considered OUTLIER");
                WOLF_TRACE("Residual: ", residual,"\nError: ",fac->getMeasurementSquareRootInformationUpper().inverse()*residual);
                remove_fac.push_back(fac_tdcp);
                // store for statistics
                outliers_tdcp_++;
                if (not sat_outliers_.count(sat))
                    sat_outliers_.emplace(sat, 1);
                else
                    sat_outliers_[sat]++;
            }
            else
                inliers_tdcp_++;
        }
    }

    // remove outliers
    for (auto fac : remove_fac)
    {
        //assert(false);
        fac->remove();
    }
    WOLF_INFO_COND(not remove_fac.empty(),
                   "ProcessorTrackerGnss::removeOutliers:",
                   "\n\tPseudorange: ", outliers_pseudorange_,
                   "\t( ", (100.0 * outliers_pseudorange_)/(outliers_pseudorange_+inliers_pseudorange_), " %)");
    if (not remove_fac.empty() and
            params_tracker_gnss_->tdcp_enabled and
            params_tracker_gnss_->remove_outliers_tdcp)
        std::cout << "\tTDCP:        "
                  << outliers_tdcp_
                  << "\t( "
                  << (100.0 * outliers_tdcp_)/(outliers_tdcp_+inliers_tdcp_)
                  << " \%)\n";
    //std::cout << "\tsats:";
    //for (auto sat_out : sat_outliers_)
    //{
    //    const int& sat = sat_out.first;
    //    int sys = satsys(sat,NULL);
    //    const unsigned int& outliers = sat_out.second;
    //    std::cout << "\t\tsat " << sat << "(" << sys << "): " << outliers << std::endl;
    //}
}

} // namespace wolf

// Register in the FactoryProcessor
#include "core/processor/factory_processor.h"
namespace wolf {
WOLF_REGISTER_PROCESSOR(ProcessorTrackerGnss)
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorTrackerGnss)
} // namespace wolf
