//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include <core/ceres_wrapper/solver_ceres.h>

#include "gnss/factor/factor_gnss_tdcp.h"
#include <core/utils/utils_gtest.h>

#include "core/problem/problem.h"
#include "gnss/sensor/sensor_gnss.h"
#include "gnss/capture/capture_gnss.h"


using namespace Eigen;
using namespace wolf;
using namespace GnssUtils;

// groundtruth transformations
Vector3d t_ecef_enu, t_enu_map, t_map_base_r, t_map_base_k, t_base_antena; // Antena extrinsics
Vector3d t_ecef_antena_r, t_ecef_antena_k;
Vector3d rpy_enu_map;
Matrix3d R_ecef_enu, R_enu_map;
Quaterniond q_map_base_r, q_map_base_k;
//double range1_r, range2_r, range3_r, range4_r, range1_k, range2_k, range3_k, range4_k;
Vector1d clock_drift_r, clock_drift_k;

GnssUtils::Satellite sat1_r({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat2_r({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat3_r({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat4_r({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat1_k({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat2_k({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat3_k({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat4_k({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Range range1_r, range2_r, range3_r, range4_r, range1_k, range2_k, range3_k, range4_k;

// WOLF
ProblemPtr prb = Problem::create("PO", 3);
SolverCeresPtr solver = std::make_shared<SolverCeres>(prb);
SensorGnssPtr gnss_sensor = std::static_pointer_cast<SensorGnss>(prb->installSensor("SensorGnss", "gnss", t_base_antena, std::make_shared<ParamsSensorGnss>()));
FrameBasePtr frm_r, frm_k;
CaptureGnssPtr cap_r, cap_k;
FeatureGnssSatellitePtr ftr1_r, ftr2_r, ftr3_r, ftr4_r, ftr1_k, ftr2_k, ftr3_k, ftr4_k;
FactorGnssTdcpPtr fac1, fac2, fac3, fac4;

void randomGroundtruth()
{
    // ECEF-ENU
    t_ecef_enu = Vector3d::Random() * 1e3;
    Vector3d rpy_ecef_enu = M_PI*Vector3d::Random();
    R_ecef_enu = (AngleAxis<double>(rpy_ecef_enu(0), Vector3d::UnitX()) *
                  AngleAxis<double>(rpy_ecef_enu(1), Vector3d::UnitY()) *
                  AngleAxis<double>(rpy_ecef_enu(2), Vector3d::UnitZ())).toRotationMatrix();

    // ENU-MAP
    t_enu_map = Vector3d::Random() * 1e3;
    rpy_enu_map = M_PI*Vector3d::Random();
    R_enu_map  = (AngleAxis<double>(rpy_enu_map(0), Vector3d::UnitX()) *
                  AngleAxis<double>(rpy_enu_map(1), Vector3d::UnitY()) *
                  AngleAxis<double>(rpy_enu_map(2), Vector3d::UnitZ())).toRotationMatrix();

    // MAP-BASE
    t_map_base_k = Vector3d::Random() * 1e2;
    q_map_base_k = Quaterniond::UnitRandom();
    t_map_base_r = Vector3d::Random() * 1e2;
    q_map_base_r = Quaterniond::UnitRandom();

    // BASE-ANTENA (EXTRINSICS)
    t_base_antena = Vector3d::Random();

    // composition
    t_ecef_antena_r = R_ecef_enu * (R_enu_map * (q_map_base_r * t_base_antena + t_map_base_r) + t_enu_map ) + t_ecef_enu;
    t_ecef_antena_k = R_ecef_enu * (R_enu_map * (q_map_base_k * t_base_antena + t_map_base_k) + t_enu_map ) + t_ecef_enu;

    // SATELLITES
    sat1_r.pos = Vector3d::Random() * 1e4;
    sat2_r.pos = Vector3d::Random() * 1e4;
    sat3_r.pos = Vector3d::Random() * 1e4;
    sat4_r.pos = Vector3d::Random() * 1e4;
    sat1_k.pos = Vector3d::Random() * 1e4;
    sat2_k.pos = Vector3d::Random() * 1e4;
    sat3_k.pos = Vector3d::Random() * 1e4;
    sat4_k.pos = Vector3d::Random() * 1e4;

    // clock drift
    clock_drift_r = Vector1d::Random() * 1e2;
    clock_drift_k = Vector1d::Random() * 1e2;

    // ranges
    range1_r.L_corrected = (sat1_r.pos-t_ecef_antena_r).norm() + clock_drift_r(0);
    range2_r.L_corrected = (sat2_r.pos-t_ecef_antena_r).norm() + clock_drift_r(0);
    range3_r.L_corrected = (sat3_r.pos-t_ecef_antena_r).norm() + clock_drift_r(0);
    range4_r.L_corrected = (sat4_r.pos-t_ecef_antena_r).norm() + clock_drift_r(0);
    range1_k.L_corrected = (sat1_k.pos-t_ecef_antena_k).norm() + clock_drift_k(0);
    range2_k.L_corrected = (sat2_k.pos-t_ecef_antena_k).norm() + clock_drift_k(0);
    range3_k.L_corrected = (sat3_k.pos-t_ecef_antena_k).norm() + clock_drift_k(0);
    range4_k.L_corrected = (sat4_k.pos-t_ecef_antena_k).norm() + clock_drift_k(0);
}

void setUpProblem()
{
    // remove previous setup
    if (frm_r)
        frm_r->remove();
    if (frm_k)
        frm_k->remove();

    solver->getSolverOptions().max_num_iterations = 100;

    // Sensor
    // ENU-MAP
    gnss_sensor->setEnuMapTranslationState(t_enu_map);
    gnss_sensor->setEnuMapRollState(rpy_enu_map(0));
    gnss_sensor->setEnuMapPitchState(rpy_enu_map(1));
    gnss_sensor->setEnuMapYawState(rpy_enu_map(2));
    // ECEF-ENU
    gnss_sensor->setEnuEcef(R_ecef_enu.transpose(), R_ecef_enu.transpose()*(-t_ecef_enu));
    // Extrinsics
    gnss_sensor->getP()->setState(t_base_antena);

    // Frame r
    Vector7d frm_r_state;
    frm_r_state.head<3>() = t_map_base_r;
    frm_r_state.tail<4>() = q_map_base_r.coeffs();
    frm_r = prb->emplaceFrame( TimeStamp(0), frm_r_state);

    // Frame k
    Vector7d frm_k_state;
    frm_k_state.head<3>() = t_map_base_k;
    frm_k_state.tail<4>() = q_map_base_k.coeffs();
    frm_k = prb->emplaceFrame( TimeStamp(1), frm_k_state);

    // capture r
    cap_r = CaptureBase::emplace<CaptureGnss>(frm_r, TimeStamp(0), gnss_sensor, nullptr);
    cap_r->getStateBlock('T')->unfix();
    cap_r->getStateBlock('T')->setState(clock_drift_r);

    // capture k
    cap_k = CaptureBase::emplace<CaptureGnss>(frm_k, TimeStamp(1), gnss_sensor, nullptr);
    cap_k->getStateBlock('T')->unfix();
    cap_k->getStateBlock('T')->setState(clock_drift_k);

    // features r
    ftr1_r = FeatureBase::emplace<FeatureGnssSatellite>(cap_r, obsd_t(), sat1_r, range1_r);
    ftr2_r = FeatureBase::emplace<FeatureGnssSatellite>(cap_r, obsd_t(), sat2_r, range2_r);
    ftr3_r = FeatureBase::emplace<FeatureGnssSatellite>(cap_r, obsd_t(), sat3_r, range3_r);
    ftr4_r = FeatureBase::emplace<FeatureGnssSatellite>(cap_r, obsd_t(), sat4_r, range4_r);

    // features k
    ftr1_k = FeatureBase::emplace<FeatureGnssSatellite>(cap_k, obsd_t(), sat1_k, range1_k);
    ftr2_k = FeatureBase::emplace<FeatureGnssSatellite>(cap_k, obsd_t(), sat2_k, range2_k);
    ftr3_k = FeatureBase::emplace<FeatureGnssSatellite>(cap_k, obsd_t(), sat3_k, range3_k);
    ftr4_k = FeatureBase::emplace<FeatureGnssSatellite>(cap_k, obsd_t(), sat4_k, range4_k);

    // factors
    fac1 = FactorBase::emplace<FactorGnssTdcp>(ftr1_r, 0.1, ftr1_r, ftr1_k, gnss_sensor, nullptr, false);
    fac2 = FactorBase::emplace<FactorGnssTdcp>(ftr2_r, 0.1, ftr2_r, ftr2_k, gnss_sensor, nullptr, false);
    fac3 = FactorBase::emplace<FactorGnssTdcp>(ftr3_r, 0.1, ftr3_r, ftr3_k, gnss_sensor, nullptr, false);
    fac4 = FactorBase::emplace<FactorGnssTdcp>(ftr4_r, 0.1, ftr4_r, ftr4_k, gnss_sensor, nullptr, false);

    // ASSERTS
    // ENU-MAP
    ASSERT_FLOAT_EQ(gnss_sensor->getEnuMapRoll()->getState()(0), rpy_enu_map(0));
    ASSERT_FLOAT_EQ(gnss_sensor->getEnuMapPitch()->getState()(0), rpy_enu_map(1));
    ASSERT_FLOAT_EQ(gnss_sensor->getEnuMapYaw()->getState()(0), rpy_enu_map(2));
    ASSERT_MATRIX_APPROX(gnss_sensor->getEnuMapTranslation()->getState(), t_enu_map, 1e-3);
    // Antena
    ASSERT_MATRIX_APPROX(gnss_sensor->getP()->getState(), t_base_antena, 1e-3);
    // Frame r
    ASSERT_MATRIX_APPROX(frm_r->getP()->getState(), t_map_base_r, 1e-3);
    ASSERT_MATRIX_APPROX(frm_r->getO()->getState(), q_map_base_r.coeffs(), 1e-3);
    // Frame k
    ASSERT_MATRIX_APPROX(frm_k->getP()->getState(), t_map_base_k, 1e-3);
    ASSERT_MATRIX_APPROX(frm_k->getO()->getState(), q_map_base_k.coeffs(), 1e-3);
    // clock drift
    ASSERT_MATRIX_APPROX(clock_drift_r, cap_r->getStateBlock('T')->getState(), 1e-3);
    ASSERT_MATRIX_APPROX(clock_drift_k, cap_k->getStateBlock('T')->getState(), 1e-3);
}

void fixAllStates()
{
    // ENU-MAP
    gnss_sensor->getEnuMapRoll()->fix();
    gnss_sensor->getEnuMapPitch()->fix();
    gnss_sensor->getEnuMapYaw()->fix();
    gnss_sensor->getEnuMapTranslation()->fix();
    // Antena
    gnss_sensor->getP()->fix();
    // Frames
    frm_r->fix();
    frm_k->fix();
    // clock drift
    cap_r->fix();
    cap_k->fix();
 }

////////////////////////////////////////////////////////
TEST(FactorGnssTdcpTest, observe_clock_drift_r)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        cap_r->getStateBlock('T')->unfix();

        // perturb
        cap_r->getStateBlock('T')->perturb(1e2);

        // Only 1 factor
        fac2->setStatus(FAC_INACTIVE);
        fac3->setStatus(FAC_INACTIVE);
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(clock_drift_r, cap_r->getStateBlock('T')->getState(), 1e-3);
    }
}

TEST(FactorGnssTdcpTest, observe_clock_drift_k)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        cap_k->getStateBlock('T')->unfix();

        // perturb
        cap_k->getStateBlock('T')->perturb(1e2);

        // Only 1 factor
        fac2->setStatus(FAC_INACTIVE);
        fac3->setStatus(FAC_INACTIVE);
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(clock_drift_k, cap_k->getStateBlock('T')->getState(), 1e-3);
    }
}

TEST(FactorGnssTdcpTest, observe_frame_p_r)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        frm_r->getP()->unfix();

        // perturb
        frm_r->getP()->perturb(1);

        // Only 3 factors
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(frm_r->getP()->getState(), t_map_base_r, 1e-3);
    }
}

TEST(FactorGnssTdcpTest, observe_frame_p_k)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        frm_k->getP()->unfix();

        // perturb
        frm_k->getP()->perturb(1);

        // Only 3 factors
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(frm_k->getP()->getState(), t_map_base_k, 1e-3);
    }
}

TEST(FactorGnssTdcpTest, observe_frame_p_clock_k)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        frm_k->getP()->unfix();
        cap_k->getStateBlock('T')->unfix();

        // perturb
        frm_k->getP()->perturb(1);
        cap_k->getStateBlock('T')->perturb(1e2);

        // all 4 factors

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(frm_k->getP()->getState(), t_map_base_k, 1e-3);
        ASSERT_MATRIX_APPROX(clock_drift_k, cap_k->getStateBlock('T')->getState(), 1e-3);
    }
}

TEST(FactorGnssTdcpTest, observe_frame_p_clock_r)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        frm_r->getP()->unfix();
        cap_r->getStateBlock('T')->unfix();

        // perturb
        frm_r->getP()->perturb(1);
        cap_r->getStateBlock('T')->perturb(1e2);

        // all 4 factors

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(frm_r->getP()->getState(), t_map_base_r, 1e-3);
        ASSERT_MATRIX_APPROX(clock_drift_r, cap_r->getStateBlock('T')->getState(), 1e-3);
    }
}

TEST(FactorGnssTdcpTest, observe_enumap_p)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        gnss_sensor->getEnuMapTranslation()->unfix();

        // perturb
        gnss_sensor->getEnuMapTranslation()->perturb(1e2);

        // Only 3 factors
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(gnss_sensor->getEnuMapTranslation()->getState(), t_enu_map, 1e-3);
    }
}

TEST(FactorGnssTdcpTest, observe_enumap_o)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        gnss_sensor->getEnuMapRoll()->unfix();
        gnss_sensor->getEnuMapPitch()->unfix();
        gnss_sensor->getEnuMapYaw()->unfix();

        // perturb
        gnss_sensor->getEnuMapRoll()->perturb(1);
        gnss_sensor->getEnuMapPitch()->perturb(1);
        gnss_sensor->getEnuMapYaw()->perturb(1);

        // Only 3 factors
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(gnss_sensor->getEnuMapTranslation()->getState(), t_enu_map, 1e-3);
    }
}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
